from enum import Enum

from django.conf import settings
from django.utils.module_loading import import_string


JOB_NAME_FUNCTION_REGISTRY = import_string(settings.JOB_NAME_FUNCTION_REGISTRY)


class JobStatusEnum(Enum):
    SUCCESS = 'S'
    FAILED = 'F'

    @classmethod
    def to_dict(cls):
        """Returns code -> description mapping for populating reference table"""
        return {status.name: status.value for status in JobStatusEnum}



from django.apps import AppConfig


JOB_STATUS_PK_MAP = {}


class JobTemplateConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'job_template'

    def ready(self):
        # Load data on startup if tables already exist (e.g., when deploying after local migrations)
        load_data_if_tables_exist()


def tables_exist(table_names):
    """Helper function to check if the tables exist in the database."""
    from django.db import connection
    with connection.cursor():
        existing_tables = connection.introspection.table_names()
        return all(table in existing_tables for table in table_names)


def load_data_if_tables_exist():
    required_tables = ['job_status_constant']
    if tables_exist(required_tables):
        load_data()


def load_data():
    global JOB_STATUS_PK_MAP
    from .models import JobStatusConstant

    JOB_STATUS_PK_MAP = {
        obj['code']: obj['id']
        for obj in JobStatusConstant.objects.values('id', 'code')
    }



import math
import traceback

from celery import shared_task
from django.utils import timezone

from .apps import JOB_STATUS_PK_MAP
from .constants import JOB_NAME_FUNCTION_REGISTRY, JobStatusEnum
from .models import JobRunMetaDataV2


@shared_task
def _execute_setup_or_context_function(job_metadata:dict, partner_id:int=None):

    records_count = additional_details = error_info = None

    task_metadata = {
        'job_name': job_metadata['job_name'],
        'job_id': job_metadata['job_id'],
        'job_run_log_id': job_metadata['job_run_log_id'],
        'config': job_metadata['config'],
        'parameters': job_metadata['parameters'],
        'partner_id': partner_id,
        'records_to_fetch_count': job_metadata['records_to_fetch_count']
    }

    # Scheduled Run
    if not job_metadata['trigger_source']:

        if setup_func := JOB_NAME_FUNCTION_REGISTRY[job_metadata['job_name']].get('setup_func'):

            setup_func_arguments = {**job_metadata['parameters']}

            # Partner level processing enabled
            if partner_id:
                setup_func_arguments['partner_id'] = partner_id

            # f(records_to_fetch_count=1000, config={}, **{})
            valid_data_list, additional_details, error_info = setup_func(
                config=job_metadata['config'],
                records_to_fetch_count = job_metadata['records_to_fetch_count'],
                **setup_func_arguments
            )

            if error_info:
                # Error encountered in setup_func, e.g. DB connection error or timeout
                if partner_id:
                    # Create an entry in JRMD
                    JobRunMetaDataV2.objects.create(
                        created_dtm=timezone.now(),
                        updated_dtm=timezone.now(),
                        job_id=job_metadata['job_id'],
                        job_run_log_id=job_metadata['job_run_log_id'],
                        metadata={'partner_id': partner_id},
                        error_info=error_info,
                        status=JOB_STATUS_PK_MAP[JobStatusEnum.FAILED.value]
                    )
                return None, None, error_info

            if valid_data_list:
                records_count = len(valid_data_list)
                if partner_id:
                    JobRunMetaDataV2.objects.create(
                        created_dtm=timezone.now(),
                        updated_dtm=timezone.now(),
                        job_id=job_metadata['job_id'],
                        job_run_log_id=job_metadata['job_run_log_id'],
                        metadata={
                            'partner_id': partner_id,
                            'eligible_records_count': records_count,
                            'additional_details': additional_details
                        },
                        status=JOB_STATUS_PK_MAP[JobStatusEnum.SUCCESS.value]
                    )

                batch_size = job_metadata['batch_size']
                if batch_size == 1:
                    for data_to_process in valid_data_list:
                        task_metadata['data'] = data_to_process
                        if isinstance(data_to_process, dict):
                            unique_id = data_to_process[job_metadata['config']['unique_key_name']]
                        else:
                             # str | int
                             unique_id = data_to_process

                        if job_metadata['run_async']:
                            try:
                                _execute_business_logic.apply_async(
                                    kwargs={
                                        'task_metadata': task_metadata
                                    },
                                    queue=job_metadata['task_queue_name']
                                )
                            except:
                                # this will happen only when Redis is down, resulting in connection error
                                JobRunMetaDataV2.objects.create(
                                    created_dtm=timezone.now(),
                                    updated_dtm=timezone.now(),
                                    job_id=job_metadata['job_id'],
                                    job_run_log_id=job_metadata['job_run_log_id'],
                                    metadata={
                                        'partner_id': partner_id,
                                        'data': unique_id
                                    },
                                    error_info=traceback.format_exc(),
                                    status=JOB_STATUS_PK_MAP[JobStatusEnum.FAILED.value]
                                )

                        else:
                            _execute_business_logic(task_metadata)

                else:
                    for idx in range(math.ceil(records_count / batch_size)):
                        start_idx = idx * batch_size
                        end_idx = start_idx + batch_size
                        batch_data_list = valid_data_list[start_idx:end_idx]
                        task_metadata['data'] = batch_data_list

                        if isinstance(batch_data_list[0], dict):
                            unique_id_list = [
                                data[job_metadata['config']['unique_key_name']]
                                for data in batch_data_list
                            ]
                        else:
                            unique_id_list = batch_data_list

                        if job_metadata['run_async']:
                            try:
                                _execute_business_logic.apply_async(
                                    kwargs={'task_metadata': task_metadata},
                                    queue=job_metadata['task_queue_name']
                                )
                            except:
                                # this will happen only when Redis is down, resulting in connection error
                                JobRunMetaDataV2.objects.create(
                                    created_dtm=timezone.now(),
                                    updated_dtm=timezone.now(),
                                    job_id=job_metadata['job_id'],
                                    job_run_log_id=job_metadata['job_run_log_id'],
                                    metadata={
                                        'partner_id': partner_id,
                                        'unique_id_list': unique_id_list
                                    },
                                    error_info=traceback.format_exc(),
                                    status=JOB_STATUS_PK_MAP[JobStatusEnum.FAILED.value]
                                )

                        else:
                            _execute_business_logic(task_metadata)

        else:
            # setup_func is not defined for the job e.g. File upload
            _execute_business_logic(task_metadata)

    # API Invocation
    else:
        _execute_business_logic(task_metadata)

    return records_count, additional_details, error_info


@shared_task
def _execute_business_logic(task_metadata:dict):

    # task_metadata format ::
    #   job_name
    #   job_id
    #   job_run_log_id
    #   config
    #   parameters
    #   partner_id
    #   data : int | str | dict

    job_run_meta_data_obj = JobRunMetaDataV2(
        created_dtm=timezone.now(),
        job_id=task_metadata['job_id'],
        job_run_log_id=task_metadata['job_run_log_id'],
        status=JOB_STATUS_PK_MAP[JobStatusEnum.SUCCESS.value]
    )

    metadata = {}
    task_metadata['config']['records_to_fetch_count'] = task_metadata['records_to_fetch_count']
    context_func_arguments = {
        'config': task_metadata['config']
    }

    # Partner level processing or passed as a parameter in case of an API
    partner_id = task_metadata.get('partner_id') or task_metadata['parameters'].get('partner_id')
    if partner_id:
        context_func_arguments['partner_id'] = partner_id
        metadata['partner_id'] = partner_id

    context_func = JOB_NAME_FUNCTION_REGISTRY[task_metadata['job_name']]['context_func']

    data_to_process = task_metadata.get('data')

    # setup_function is defined
    if data_to_process:
        # str | int
        unique_id_or_id_list = data_to_process
        if isinstance(data_to_process, dict):
            unique_id_or_id_list = data_to_process[task_metadata['config']['unique_key_name']]
        elif isinstance(data_to_process, list) and isinstance(data_to_process[0], dict):
            unique_id_or_id_list = [
                data[task_metadata['config']['unique_key_name']]
                for data in data_to_process
            ]
        metadata['unique_id_or_id_list'] = unique_id_or_id_list
        error_info = context_func(data_to_process, **context_func_arguments)
    else:
        # when setup_function is not defined
        metadata['parameters'] = task_metadata['parameters']
        error_info = context_func(**task_metadata['parameters'], **context_func_arguments)

    if error_info:
        job_run_meta_data_obj.status = JOB_STATUS_PK_MAP[JobStatusEnum.FAILED.value]

    job_run_meta_data_obj.metadata = metadata
    job_run_meta_data_obj.error_info = error_info
    job_run_meta_data_obj.updated_dtm = timezone.now()
    job_run_meta_data_obj.save()



from django.conf import settings
from django.db import models


class DefaultTimeStamp(models.Model):
    created_dtm = models.DateTimeField(auto_now_add=True)
    updated_dtm = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class JobTypeEnum(models.TextChoices):
    PARTNER = 'P', 'PARTNER'
    SYSTEM = 'SY', 'SYSTEM'


class JobConfiguration(DefaultTimeStamp):
    job_name = models.CharField(max_length=100)
    is_scheduled_run_enabled = models.BooleanField(default=True)
    is_realtime_run_enabled = models.BooleanField(default=False)
    batch_size = models.PositiveSmallIntegerField()
    records_to_fetch_count = models.PositiveIntegerField()
    # this is to handle cases when we run jobs locally and want to execute batches sequentially or w/o celery
    run_async = models.BooleanField(default=False)
    task_queue_name = models.CharField(max_length=30, default=settings.CELERY_TASK_DEFAULT_QUEUE)
    process_at_partner_level = models.BooleanField(default=False)
    config = models.JSONField(null=True)
    type = models.CharField(max_length=2, choices=JobTypeEnum.choices, default=JobTypeEnum.PARTNER)

    class Meta:
        db_table = 'job_configuration'
        verbose_name = 'Job Configuration'
        verbose_name_plural = 'Job Configurations'


class JobPartnerMapping(DefaultTimeStamp):
    job = models.ForeignKey(JobConfiguration, related_name='partner_configs', on_delete=models.DO_NOTHING)
    partner_id = models.PositiveSmallIntegerField()
    is_enabled = models.BooleanField(default=False)
    config = models.JSONField(null=True)

    class Meta:
        db_table = 'job_partner_mapping'
        unique_together = ('job', 'partner_id')
        verbose_name = 'Job Partner Mapping'
        verbose_name_plural = 'Job Partner Mappings'


class PeriodicJobRunLog(models.Model):
    created_dtm = models.DateTimeField()
    updated_dtm = models.DateTimeField()
    job_id = models.PositiveSmallIntegerField()
    # 32-character lowercase hexadecimal string
    job_run_log_id = models.CharField(max_length=35)
    status = models.PositiveSmallIntegerField()
    parameters = models.JSONField(null=True)
    # this will be populated only if setup_func fails or push to redis fails
    error_info = models.TextField(null=True)

    class Meta:
        db_table = 'periodic_job_run_log'
        verbose_name = 'Periodic Job Run Log'
        verbose_name_plural = 'Periodic Job Run Logs'


class JobRunMetaDataV2(models.Model):
    created_dtm = models.DateTimeField()
    updated_dtm = models.DateTimeField()
    job_id = models.PositiveSmallIntegerField()
    job_run_log_id = models.CharField(max_length=35, null=True)
    metadata = models.JSONField(null=True)
    error_info = models.TextField(null=True)
    status = models.PositiveSmallIntegerField()

    class Meta:
        db_table = 'job_run_meta_data_v2'
        verbose_name = 'Job Run Meta Data V2'
        verbose_name_plural = 'Job Run Meta Data V2'


class JobStatusConstant(DefaultTimeStamp):
    status = models.CharField(max_length=10)
    code = models.CharField(max_length=2)

    class Meta:
        db_table = 'job_status_constant'
        verbose_name = 'Job Status Constant'
        verbose_name_plural = 'Job Status Constants'



# Generated by Django 5.1.5 on 2025-09-24 16:21

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='JobConfiguration',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_dtm', models.DateTimeField(auto_now_add=True)),
                ('updated_dtm', models.DateTimeField(auto_now=True)),
                ('job_name', models.CharField(max_length=100)),
                ('is_scheduled_run_enabled', models.BooleanField(default=True)),
                ('is_realtime_run_enabled', models.BooleanField(default=False)),
                ('batch_size', models.PositiveSmallIntegerField()),
                ('records_to_fetch_count', models.PositiveIntegerField()),
                ('run_async', models.BooleanField(default=False)),
                ('task_queue_name', models.CharField(default='celery', max_length=30)),
                ('process_at_partner_level', models.BooleanField(default=False)),
                ('config', models.JSONField(null=True)),
                ('type', models.CharField(choices=[('P', 'PARTNER'), ('SY', 'SYSTEM')], default='P', max_length=2)),
            ],
            options={
                'verbose_name': 'Job Configuration',
                'verbose_name_plural': 'Job Configurations',
                'db_table': 'job_configuration',
            },
        ),
        migrations.CreateModel(
            name='JobRunMetaDataV2',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_dtm', models.DateTimeField()),
                ('updated_dtm', models.DateTimeField()),
                ('job_id', models.PositiveSmallIntegerField()),
                ('job_run_log_id', models.CharField(max_length=35, null=True)),
                ('metadata', models.JSONField(null=True)),
                ('error_info', models.TextField(null=True)),
                ('status', models.PositiveSmallIntegerField()),
            ],
            options={
                'verbose_name': 'Job Run Meta Data V2',
                'verbose_name_plural': 'Job Run Meta Data V2',
                'db_table': 'job_run_meta_data_v2',
            },
        ),
        migrations.CreateModel(
            name='JobStatusConstant',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_dtm', models.DateTimeField(auto_now_add=True)),
                ('updated_dtm', models.DateTimeField(auto_now=True)),
                ('status', models.CharField(max_length=10)),
                ('code', models.CharField(max_length=2)),
            ],
            options={
                'verbose_name': 'Job Status Constant',
                'verbose_name_plural': 'Job Status Constants',
                'db_table': 'job_status_constant',
            },
        ),
        migrations.CreateModel(
            name='PeriodicJobRunLog',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_dtm', models.DateTimeField()),
                ('updated_dtm', models.DateTimeField()),
                ('job_id', models.PositiveSmallIntegerField()),
                ('job_run_log_id', models.CharField(max_length=35)),
                ('status', models.PositiveSmallIntegerField()),
                ('parameters', models.JSONField(null=True)),
                ('error_info', models.TextField(null=True)),
            ],
            options={
                'verbose_name': 'Periodic Job Run Log',
                'verbose_name_plural': 'Periodic Job Run Logs',
                'db_table': 'periodic_job_run_log',
            },
        ),
        migrations.CreateModel(
            name='JobPartnerMapping',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_dtm', models.DateTimeField(auto_now_add=True)),
                ('updated_dtm', models.DateTimeField(auto_now=True)),
                ('partner_id', models.PositiveSmallIntegerField()),
                ('is_enabled', models.BooleanField(default=False)),
                ('config', models.JSONField(null=True)),
                ('job', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='partner_configs', to='job_template.jobconfiguration')),
            ],
            options={
                'verbose_name': 'Job Partner Mapping',
                'verbose_name_plural': 'Job Partner Mappings',
                'db_table': 'job_partner_mapping',
                'unique_together': {('job', 'partner_id')},
            },
        ),
    ]



import json
import traceback

from django.db.models import Prefetch

from .models import JobConfiguration, JobPartnerMapping, JobTypeEnum
from .tasks import _execute_setup_or_context_function


def check_job_status_and_fetch_config(job_name, trigger_source, job_type):
    filters = {
        'job_name': job_name,
        'type': job_type
    }

    if trigger_source:
        filters['is_realtime_run_enabled'] = True
    else:
        filters['is_scheduled_run_enabled'] = True

    job_config_qs = JobConfiguration.objects.filter(
        **filters
    ).only(
        'id', 'batch_size', 'records_to_fetch_count', 'task_queue_name', 'process_at_partner_level', 'config',
        'run_async'
    )

    if job_type == JobTypeEnum.PARTNER:
        job_config_qs = job_config_qs.prefetch_related(
            Prefetch(
                'partner_configs',
                queryset=JobPartnerMapping.objects.filter(is_enabled=True).only('partner_id', 'config'),
                to_attr='active_partner_configs'
            )
        )

    job_config_data = job_config_qs[:1]

    if job_config_data:
        job_config_obj = job_config_data[0]
        job_config = {
            'job_name': job_name,
            'trigger_source': trigger_source,
            'job_id':  job_config_obj.id,
            'batch_size': job_config_obj.batch_size,
            'records_to_fetch_count': job_config_obj.records_to_fetch_count,
            'run_async': job_config_obj.run_async,
            'task_queue_name': job_config_obj.task_queue_name,
            'process_at_partner_level': job_config_obj.process_at_partner_level,
            'config': job_config_obj.config or {},
            'job_type': job_type
        }

        if job_type == JobTypeEnum.PARTNER:
            if not job_config_obj.active_partner_configs:
                return None
            else:
                job_config['partner_config_map'] = {
                    partner_config_obj.partner_id: partner_config_obj.config or {}
                    for partner_config_obj in job_config_obj.active_partner_configs
                }

        return job_config

    # Even if Job is enabled, there can be no partners mapped to it
    return None


def _process_task(job_metadata: dict, config:dict) -> tuple[int | None, dict | None, str | None]:

    records_count = additional_details = error_info = None
    error_info_map = {}

    if job_metadata['process_at_partner_level']:

        partner_config_map = config.pop('partner_config_map')
        allowed_partner_id_set = partner_config_map.keys()

        # if partner_id is passed, then we need to run only for given partners else all partners from the config
        if partner_id_list := job_metadata['partner_id_list']:
            partner_id_set = set(partner_id_list)
            diff = partner_id_set - allowed_partner_id_set
            if diff:
                # in this case, neither PJRL nor JRMD will be created if it's not a scheduled run
                return None, None, f'Either partner is disabled or Config missing for partner_id(s) : {diff}'
        else:
            partner_id_set = allowed_partner_id_set

        run_async = job_metadata['run_async']
        for partner_id in partner_id_set:
            job_metadata['config'] = {
                **config,
                **partner_config_map[partner_id]
            }
            if run_async:
                try:
                    _execute_setup_or_context_function.apply_async(
                        kwargs={
                            'job_metadata': job_metadata,
                            'partner_id': partner_id
                        },
                        queue=job_metadata['task_queue_name']
                    )
                except:
                    # Redis connection error
                    error_info_map[partner_id] = traceback.format_exc()
            else:
                _execute_setup_or_context_function(job_metadata, partner_id)

    else:
        job_metadata['config'] = config
        records_count, additional_details, error_info = _execute_setup_or_context_function(job_metadata)

    if error_info_map:
        error_info = json.dumps(error_info_map)

    return records_count, additional_details, error_info



from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from job_template.constants import JobStatusEnum
from job_template.models import JobStatusConstant


class Command(BaseCommand):
    help = 'Populate JobStatusConstant table with status and corresponding code'

    def add_arguments(self, parser):
        parser.add_argument(
            '--clear',
            action='store_true',
            help='Clear existing status constants before populating',
        )
        parser.add_argument(
            '--dry-run',
            action='store_true',
            help='Show what would be created without actually creating',
        )

    def handle(self, *args, **options):
        status_mappings = JobStatusEnum.to_dict()

        if options['dry_run']:
            self.stdout.write(
                self.style.WARNING('DRY RUN - No changes will be made')
            )
            self.stdout.write('Would create/update the following status constants:')
            for status, code in status_mappings.items():
                self.stdout.write(f'{code} -> {status}')
            return

        try:
            with transaction.atomic():
                if options['clear']:
                    deleted_count, _ = JobStatusConstant.objects.all().delete()
                    self.stdout.write(
                        self.style.WARNING(f'Cleared {deleted_count} existing status constants')
                    )

                created_count = 0
                updated_count = 0

                for status, code in status_mappings.items():
                    obj, created = JobStatusConstant.objects.update_or_create(
                        status=status,
                        defaults={'code': code}
                    )
                    
                    if created:
                        created_count += 1
                        self.stdout.write(f'Created: {code} -> {status}')
                    else:
                        updated_count += 1
                        self.stdout.write(f'Updated: {code} -> {status}')

                self.stdout.write(
                    self.style.SUCCESS(
                        f'Successfully processed {len(status_mappings)} status constants '
                        f'({created_count} created, {updated_count} updated)'
                    )
                )

        except Exception as e:
            raise CommandError(f'Error populating status constants: {str(e)}')



import uuid

from django.utils import timezone

from .apps import JOB_STATUS_PK_MAP
from .constants import JobStatusEnum
from .models import JobTypeEnum, PeriodicJobRunLog
from .utils import _process_task


class JobTemplate:

    def __init__(self, *, job_config:dict, partner_id_list:list[int]=None, **kwargs:dict):
        self.config = job_config.pop('config')
        self.job_metadata = job_config
        self.job_metadata['parameters'] = kwargs
        self.job_metadata['partner_id_list'] = partner_id_list
        self.job_parameters = {
            'parameters': kwargs,
            'partner_id_list': partner_id_list,
        }

        job_type = job_config.pop('job_type')
        if job_type == JobTypeEnum.PARTNER:
            self.config['partner_config_map'] = job_config.pop('partner_config_map')
            self.job_parameters['active_partner_id_list'] = list(self.config['partner_config_map'].keys())

    def execute_job(self):
        self.job_metadata['job_run_log_id'] = uuid.uuid4().hex
        is_scheduled_run = not self.job_metadata['trigger_source']

        # We won't be able to handle job invocation using API, since in that scenario is_scheduled_run will evaluate
        # to True and both setup & context function will be invoked. But by looking at PJRL, we won't be able to
        # decide how the job executed before its schedule. Hence for API invocation, trigger_source and pk_list needs
        # to be passed all the time.
        if is_scheduled_run:
            periodic_job_run_log_obj = PeriodicJobRunLog(
                created_dtm=timezone.now(),
                job_id=self.job_metadata['job_id'],
                job_run_log_id=self.job_metadata['job_run_log_id'],
                status=JOB_STATUS_PK_MAP[JobStatusEnum.SUCCESS.value]
            )

        records_count, additional_details, error_info = _process_task(self.job_metadata, self.config)

        if is_scheduled_run:
            self.job_parameters['eligible_records_count'] = records_count

            if error_info:
                periodic_job_run_log_obj.status = JOB_STATUS_PK_MAP[JobStatusEnum.FAILED.value]

            if additional_details:
                self.job_parameters.update(additional_details)

            periodic_job_run_log_obj.parameters = self.job_parameters
            periodic_job_run_log_obj.error_info = error_info
            periodic_job_run_log_obj.updated_dtm = timezone.now()
            periodic_job_run_log_obj.save()
